//------------------------------------------------------------------------------
// This file is part of XrdHTTP: A pragmatic implementation of the
// HTTP/WebDAV protocol for the Xrootd framework
//
// Copyright (c) 2013 by European Organization for Nuclear Research (CERN)
// Author: Fabrizio Furano <furano@cern.ch>
// File Date: Nov 2012
//------------------------------------------------------------------------------
// XRootD is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// XRootD is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with XRootD.  If not, see <http://www.gnu.org/licenses/>.
//------------------------------------------------------------------------------









/** @file  XrdHttpStatic.hh
 * @brief  Static resources, here for performance and ease of setup
 * @author Fabrizio Furano
 * @date   Jun 2013
 * 
 * 
 * 
 */



#ifndef XRDHTTPSTATIC_HH
#define	XRDHTTPSTATIC_HH




// Small static resources that for practical and performance reasons we want
// to keep here. In general these are obtained with "xdd -i <filename>"

#include "static/xrdhttp_css.h"
#include "static/xrdhttp_favicon_ico.h"
//#include "static/xrdhttp_logo.h"


#endif	/* XRDHTTPSTATIC_HH */

