#ifndef _XRDTLS_TRACE_H
#define _XRDTLS_TRACE_H
/******************************************************************************/
/*                                                                            */
/*                        X r d T l s T r a c e . h h                         */
/*                                                                            */
/* (c) 2020 by the Board of Trustees of the Leland Stanford, Jr., University  */
/*                            All Rights Reserved                             */
/*   Produced by Andrew Hanushevsky for Stanford University under contract    */
/*              DE-AC02-76-SFO0515 with the Department of Energy              */
/*                                                                            */
/* This file is part of the XRootD software suite.                            */
/*                                                                            */
/* XRootD is free software: you can redistribute it and/or modify it under    */
/* the terms of the GNU Lesser General Public License as published by the     */
/* Free Software Foundation, either version 3 of the License, or (at your     */
/* option) any later version.                                                 */
/*                                                                            */
/* XRootD is distributed in the hope that it will be useful, but WITHOUT      */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or      */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public       */
/* License for more details.                                                  */
/*                                                                            */
/* You should have received a copy of the GNU Lesser General Public License   */
/* along with XRootD in a file called COPYING.LESSER (LGPL license) and file  */
/* COPYING (GPL license).  If not, see <http://www.gnu.org/licenses/>.        */
/*                                                                            */
/* The copyright holder's institutional names and contributor's names may not */
/* be used to endorse or promote products derived from this software without  */
/* specific prior written permission of the institution or contributor.       */
/******************************************************************************/

#ifndef NODEBUG

#include "XrdSys/XrdSysTrace.hh"

#define TRACING(act) XrdTlsGlobal::SysTrace.What & act

#define DBG_CTX(y)\
    if (XrdTlsGlobal::SysTrace.What & XrdTls::dbgCTX) \
       SYSTRACE(XrdTlsGlobal::SysTrace., 0, epname, 0, y)

#define DBG_SOK(y)\
    if (XrdTlsGlobal::SysTrace.What & XrdTls::dbgSOK) \
       SYSTRACE(XrdTlsGlobal::SysTrace., pImpl->traceID, epname, 0, y)

#define DBG_SIO(y)\
    if (XrdTlsGlobal::SysTrace.What & XrdTls::dbgSIO) \
       SYSTRACE(XrdTlsGlobal::SysTrace., pImpl->traceID, epname, 0, y)

#define DBG_TLS(y)\
       SYSTRACE(XrdTlsGlobal::SysTrace., pImpl->traceID, epname, 0, y)

#define EPNAME(x) static const char *epname = x;

#else

#define DBG_CTX(y)
#define DBG_SOK(y)
#define DBG_SIO(y)
#define EPNAME(x)
#define QTRACE(act) false

#endif
#endif
